from __future__ import annotations

import argparse
import shutil
import sys
from pathlib import Path
from typing import Callable, Dict


Handler = Callable[[], int]


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="homa",
        description="Command-line helpers for the homa package.",
    )
    parser.add_argument(
        "command",
        help="Name of the command to run, e.g. cache:remove.",
    )
    return parser


def _command_table() -> Dict[str, Handler]:
    return {
        "cache:remove": _handle_cache_remove,
    }


def _handle_cache_remove() -> int:
    root = Path.cwd()
    removed = 0
    errors: list[str] = []

    for candidate in root.rglob("__pycache__"):
        if not candidate.is_dir():
            continue
        try:
            shutil.rmtree(candidate)
            removed += 1
        except OSError as exc:
            errors.append(f"{candidate}: {exc}")

    print(f"Removed {removed} __pycache__ director{'ies' if removed != 1 else 'y'}.")

    if errors:
        print("Failed to remove the following paths:", file=sys.stderr)
        for error in errors:
            print(f"  - {error}", file=sys.stderr)
        return 1

    return 0


def main(argv: list[str] | None = None) -> int:
    parser = _build_parser()
    args = parser.parse_args(argv)

    commands = _command_table()
    handler = commands.get(args.command)
    if handler is None:
        parser.error(f"Unknown command '{args.command}'.")
        return 2

    return handler()


if __name__ == "__main__":
    sys.exit(main())



import torch
from torchvision.models import resnet50
from .utils import replace_relu
from ..activation import StochasticActivation


class StochasticResnet(torch.nn.Module):
    def __init__(self, outputs: int):
        super(StochasticResnet, self).__init__()
        self._create_encoder()
        self.fc = torch.nn.Linear(2048, outputs)
        replace_relu(self.resnet, StochasticActivation)

    def _create_encoder(self):
        self.encoder = resnet50(weights="DEFAULT")
        self.encoder.fc = torch.nn.Identity()

    def forward(self, images: torch.Tensor):
        features = self.encoder(images)
        return self.fc(features)



import torch


class GALU(torch.nn.Module):
    def __init__(self, max_input: float = 1.0):
        super(GALU, self).__init__()
        if max_input <= 0:
            raise ValueError("max_input must be positive.")
        self.max_input = max_input
        self.alpha = None
        self.beta = None
        self.gamma = None
        self.delta = None
        self._num_channels = None

    def _initialize_parameters(self, x):
        if x.ndim < 2:
            raise ValueError(
                f"Input tensor must have at least 2 dimensions (N, C), but got shape {x.shape}"
            )

        num_channels = x.shape[1]
        self._num_channels = num_channels
        param_shape = [1] * x.ndim
        param_shape[1] = num_channels
        self.alpha = torch.nn.Parameter(torch.zeros(param_shape))
        self.beta = torch.nn.Parameter(torch.zeros(param_shape))
        self.gamma = torch.nn.Parameter(torch.zeros(param_shape))
        self.delta = torch.nn.Parameter(torch.zeros(param_shape))

    def forward(self, x):
        if self.alpha is None:
            self._initialize_parameters(x)

        zero = torch.tensor(0.0, device=x.device, dtype=x.dtype)
        x_norm = x / self.max_input
        part_prelu = torch.relu(x_norm) + self.alpha * torch.min(x_norm, zero)
        part_beta = self.beta * (
            torch.relu(1.0 - torch.abs(x_norm - 1.0))
            + torch.min(torch.abs(x_norm - 3.0) - 1.0, zero)
        )
        part_gamma = self.gamma * (
            torch.relu(0.5 - torch.abs(x_norm - 0.5))
            + torch.min(torch.abs(x_norm - 1.5) - 0.5, zero)
        )
        part_delta = self.delta * (
            torch.relu(0.5 - torch.abs(x_norm - 2.5))
            + torch.min(torch.abs(x_norm - 3.5) - 0.5, zero)
        )
        z = part_prelu + part_beta + part_gamma + part_delta
        return z * self.max_input



from .APLU import APLU
from .GALU import GALU
from .SmallGALU import SmallGALU
from .MELU import MELU
from .WideMELU import WideMELU
from .PDELU import PDELU
from .SReLU import SReLU
from .StochasticActivation import StochasticActivation



import torch


def replace_relu(model: torch.nn.Module, activation: torch.nn.Module) -> int:
    replaced = 0
    for parent in model.modules():
        for name, child in list(parent.named_children()):
            if isinstance(child, torch.nn.ReLU):
                setattr(parent, name, activation())
                replaced += 1
    return replaced



import torch


class APLU(torch.nn.Module):
    def __init__(self, max_input: float = 1.0):
        super(APLU, self).__init__()
        self.max_input = max_input
        self.alpha = None
        self.beta = None
        self.gamma = None
        self.xi = None
        self.psi = None
        self.mu = None
        self._num_channels = None

    def _initialize_parameters(self, x):
        if x.ndim < 2:
            raise ValueError(
                f"Input tensor must have at least 2 dimensions (N, C), but got shape {x.shape}"
            )

        num_channels = x.shape[1]
        self._num_channels = num_channels

        param_shape = [1] * x.ndim
        param_shape[1] = num_channels

        self.alpha = torch.nn.Parameter(torch.zeros(param_shape))
        self.beta = torch.nn.Parameter(torch.zeros(param_shape))
        self.gamma = torch.nn.Parameter(torch.zeros(param_shape))

        self.xi = torch.nn.Parameter(self.max_input * torch.rand(param_shape))
        self.psi = torch.nn.Parameter(self.max_input * torch.rand(param_shape))
        self.mu = torch.nn.Parameter(self.max_input * torch.rand(param_shape))

    def forward(self, x):
        if self.alpha is None:
            self._initialize_parameters(x)

        a = torch.relu(x)

        # following are called hinges
        b = self.alpha * torch.relu(-x + self.xi)
        c = self.beta * torch.relu(-x + self.psi)
        d = self.gamma * torch.relu(-x + self.mu)
        z = a + b + c + d

        return z



import torch


class MELU(torch.nn.Module):
    def __init__(self, maxInput: float = 1.0):
        super().__init__()
        self.maxInput = float(maxInput)
        self.alpha = None
        self.beta = None
        self.gamma = None
        self.delta = None
        self.xi = None
        self.psi = None
        self._initialized = False

    def _initialize_parameters(self, X: torch.Tensor):
        if X.dim() != 4:
            raise ValueError(
                f"Expected 4D input (B, C, H, W), but got {X.dim()}D input."
            )
        num_channels = X.shape[1]
        shape = (1, num_channels, 1, 1)
        self.alpha = torch.nn.Parameter(torch.zeros(shape))
        self.beta = torch.nn.Parameter(torch.zeros(shape))
        self.gamma = torch.nn.Parameter(torch.zeros(shape))
        self.delta = torch.nn.Parameter(torch.zeros(shape))
        self.xi = torch.nn.Parameter(torch.zeros(shape))
        self.psi = torch.nn.Parameter(torch.zeros(shape))
        self._initialized = True

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        if not self._initialized:
            self._initialize_parameters(X)
        X_norm = X / self.maxInput
        Y = torch.roll(X_norm, shifts=-1, dims=1)
        term1 = torch.relu(X_norm)
        term2 = self.alpha * torch.clamp(X_norm, max=0)
        dist_sq_beta = (X_norm - 2) ** 2 + (Y - 2) ** 2
        dist_sq_gamma = (X_norm - 1) ** 2 + (Y - 1) ** 2
        dist_sq_delta = (X_norm - 1) ** 2 + (Y - 3) ** 2
        dist_sq_xi = (X_norm - 3) ** 2 + (Y - 1) ** 2
        dist_sq_psi = (X_norm - 3) ** 2 + (Y - 3) ** 2
        term3 = self.beta * torch.sqrt(torch.relu(2 - dist_sq_beta))
        term4 = self.gamma * torch.sqrt(torch.relu(1 - dist_sq_gamma))
        term5 = self.delta * torch.sqrt(torch.relu(1 - dist_sq_delta))
        term6 = self.xi * torch.sqrt(torch.relu(1 - dist_sq_xi))
        term7 = self.psi * torch.sqrt(torch.relu(1 - dist_sq_psi))
        Z_norm = term1 + term2 + term3 + term4 + term5 + term6 + term7
        Z = Z_norm * self.maxInput
        return Z
